----------------
-----Models-----
----------------

-- If any of these are needed in another file, like the exclamation box, just make models you need global

local E_MODEL_OW1 = smlua_model_util_get_id("ow1_geo")
local E_MODEL_C2 = smlua_model_util_get_id("c2_geo")
local E_MODEL_C8PLAT = smlua_model_util_get_id("c8plat_geo")
local E_MODEL_C9PLAT = smlua_model_util_get_id("c9plat_geo")
local E_MODEL_STATICPLAT = smlua_model_util_get_id("staticplat_geo")
local E_MODEL_WDWPLAT = smlua_model_util_get_id("wdwplat_geo")
local E_MODEL_C12PLAT = smlua_model_util_get_id("c12log_geo")

--------------------
-----Collisions-----
--------------------

local COL_OW1 = smlua_collision_util_get("ow1_collision")
local COL_C2 = smlua_collision_util_get("c2_collision")
local COL_C3 = smlua_collision_util_get("ssl_seg7_collision_grindel")
local COL_C4 = smlua_collision_util_get("wdw_seg7_collision_express_elevator_platform")
local COL_C5DO = smlua_collision_util_get("rr_seg7_collision_donut_platform")
local COL_C8PLAT = smlua_collision_util_get("c8plat_collision")
local COL_C9PLAT = smlua_collision_util_get("c9plat_collision")
local COL_STATICPLAT = smlua_collision_util_get("staticplat_collision")
local COL_WDWPLAT = smlua_collision_util_get("wdwplat_collision")
local COL_C12PLAT = smlua_collision_util_get("c12log_collision")

-----------------------------
-----Localized Functions-----
-----------------------------

local spawn_non_sync_object = spawn_non_sync_object
local obj_set_model_extended = obj_set_model_extended
local network_init_object = network_init_object
local load_object_collision_model = load_object_collision_model
local cur_obj_become_tangible = cur_obj_become_tangible
local obj_set_hitbox_radius_and_height = obj_set_hitbox_radius_and_height
local obj_mark_for_deletion = obj_mark_for_deletion
local obj_scale = obj_scale
local cur_obj_scale = cur_obj_scale

-------------------
------Helpers------
-------------------

---@param parent Object
---@param model ModelExtendedId
---@param behaviorId BehaviorId
local function spawn_object(parent, model, behaviorId)
	local obj = spawn_non_sync_object(behaviorId, model, 0, 0, 0, nil)
	if not obj then return nil end

	obj_copy_pos_and_angle(obj, parent)
	return obj
end

function get_star_count()
	return save_file_get_total_star_count
	(get_current_save_file_num() - 1, COURSE_NONE, COURSE_SA)
end


--- @param obj Object
--- @param hitbox ObjectHitbox
function obj_set_hitbox(obj, hitbox)
	if not obj or not hitbox then return end
	-- Sets other hitbox values once
	if (obj.oFlags & OBJ_FLAG_30) == 0 then
		obj.oFlags = obj.oFlags | OBJ_FLAG_30

		obj.oInteractType = hitbox.interactType
		obj.oDamageOrCoinValue = hitbox.damageOrCoinValue
		obj.oHealth = hitbox.health
		obj.oNumLootCoins = hitbox.numLootCoins

		cur_obj_become_tangible()
	end

	-- Set actual hitboxes
	obj.hitboxRadius = obj.header.gfx.scale.x * hitbox.radius
	obj.hitboxHeight = obj.header.gfx.scale.y * hitbox.height
	obj.hurtboxRadius = obj.header.gfx.scale.x * hitbox.hurtboxRadius
	obj.hurtboxHeight = obj.header.gfx.scale.y * hitbox.hurtboxHeight
	obj.hitboxDownOffset = obj.header.gfx.scale.y * hitbox.downOffset
end

-------------------
-----Functions-----
-------------------

------ OW1 Wall ------

---@param o Object
local function ow1wall_init(o)
	obj_set_model_extended(o, E_MODEL_OW1)
	o.collisionData = COL_OW1
	o.oCollisionDistance = 24582
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
	obj_set_hitbox_radius_and_height(o, 596, 1680)
end

local function ow1wall_loop()
	bhv_wf_breakable_wall_loop()
	load_object_collision_model()
end

hook_behavior(id_bhvCcmTouchedStarSpawn, OBJ_LIST_SURFACE, true, ow1wall_init, ow1wall_loop)

------ C2 PLAT ------

---@param o Object
local function c2rot_init(o)
	obj_set_model_extended(o, E_MODEL_C2)
	o.collisionData = COL_C2
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
	network_init_object(o, true, { "oFaceAngleYaw", "oAngleVelYaw" })
end

---@param o Object
local function c2rot_loop(o)
	o.oFaceAngleYaw = o.oFaceAngleYaw + 512
	o.oAngleVelYaw = 512
	load_object_collision_model()
end

hook_behavior(id_bhvControllablePlatform, OBJ_LIST_SURFACE, true, c2rot_init, c2rot_loop)

------ C3 GRINDEL ------

---@param o Object
local function gri_init(o)
	obj_set_model_extended(o, E_MODEL_SSL_GRINDEL)
	o.collisionData = COL_C3
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
end

local function gri_loop()
	load_object_collision_model()
end

hook_behavior(id_bhvPyramidTopFragment, OBJ_LIST_SURFACE, true, gri_init, gri_loop)

------ C4 PLATFORM ------

---@param o Object
local function c4plat_init(o)
	o.collisionData = COL_C4
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
end

local function c4plat_loop()
	bhv_ssl_moving_pyramid_wall_loop()
	load_object_collision_model()
end

hook_behavior(id_bhvPenguinRaceShortcutCheck, OBJ_LIST_SURFACE, true, c4plat_init, c4plat_loop)

------ C5 DONUT ------

---@param o Object
local function c5donut_init(o)
	o.collisionData = COL_C5DO
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
end

local function c5donut_loop()
	bhv_tumbling_bridge_platform_loop()
	load_object_collision_model()
end

hook_behavior(id_bhvDonutPlatformSpawner, OBJ_LIST_SURFACE, true, c5donut_init, c5donut_loop)

------ C8 PLAT ------

---@param o Object
local function c8plat_init(o)
	obj_set_model_extended(o, E_MODEL_C8PLAT)
	o.collisionData = COL_C8PLAT
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
end

local function c8plat_loop()
	bhv_tilting_inverted_pyramid_loop()
	load_object_collision_model()
end

hook_behavior(id_bhvMeshElevator, OBJ_LIST_SURFACE, true, c8plat_init, c8plat_loop)

------ C9 PLAT ------

---@param o Object
local function c9plat_init(o)
	obj_set_model_extended(o, E_MODEL_C9PLAT)
	o.collisionData = COL_C9PLAT
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
	network_init_object(o, true, { "oFaceAngleYaw", "oAngleVelYaw" })
end

---@param o Object
local function c9plat_loop(o)
	o.oFaceAngleYaw = o.oFaceAngleYaw + 256
	o.oAngleVelYaw = 256
	load_object_collision_model()
end

hook_behavior(id_bhvLllMovingOctagonalMeshPlatform, OBJ_LIST_SURFACE, true, c9plat_init, c9plat_loop)

------ C9 STATICPLAT ------

---@param o Object
local function c9plat2_init(o)
	obj_set_model_extended(o, E_MODEL_STATICPLAT)
	o.collisionData = COL_STATICPLAT
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
end

local function c9plat2_loop()
	load_object_collision_model()
end

hook_behavior(id_bhvLllHexagonalMesh, OBJ_LIST_SURFACE, true, c9plat2_init, c9plat2_loop)

------ WDWPLAT ------

---@param o Object
local function wdwplat_init(o)
	obj_set_model_extended(o, E_MODEL_WDWPLAT)
	o.collisionData = COL_WDWPLAT
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
end

local function wdwplat_loop()
	bhv_lll_drawbridge_loop()
	load_object_collision_model()
end

hook_behavior(id_bhvPyramidElevator, OBJ_LIST_SURFACE, true, wdwplat_init, wdwplat_loop)

------ C12PLAT ------

---@param o Object
local function c12plat_init(o)
	obj_set_model_extended(o, E_MODEL_C12PLAT)
	o.collisionData = COL_C12PLAT
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
	o.oCollisionDistance = 12336
	network_init_object(o, true, { "oFaceAngleYaw", "oAngleVelYaw" })
end

---@param o Object
local function c12plat_loop(o)
	o.oFaceAngleYaw = o.oFaceAngleYaw + 256
	o.oAngleVelYaw = 256
	load_object_collision_model()
end

hook_behavior(id_bhvFallingPillar, OBJ_LIST_SURFACE, true, c12plat_init, c12plat_loop)

------ C12PLAT BASE ------

---@param o Object
local function c12platbase_init(o)
	obj_set_model_extended(o, E_MODEL_C12PLAT)
	o.collisionData = COL_C12PLAT
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	o.header.gfx.skipInViewCheck = true
	o.oCollisionDistance = 12336
	network_init_object(o, true, { "oFaceAnglePitch", "oAngleVelPitch" })
end

---@param o Object
local function c12platbase_loop(o)
	o.oFaceAnglePitch = o.oFaceAnglePitch + 256
	o.oAngleVelPitch = 256
	load_object_collision_model()
end

hook_behavior(id_bhvPillarBase, OBJ_LIST_SURFACE, true, c12platbase_init, c12platbase_loop)

------ SECRETS ------

local sSecret = {
interactType      = INTERACT_COIN,
downOffset        = 0,
damageOrCoinValue = 0,
health            = 0,
numLootCoins      = 0,
radius            = 90,
height            = 90,
hurtboxRadius     = 0,
hurtboxHeight     = 0,
}

---@param o Object
local function starsecret_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    obj_set_hitbox(o, sSecret)
    obj_set_model_extended(o, E_MODEL_STAR)
end

---@param o Object
local function starsecret_loop(o)
    o.oFaceAngleYaw = o.oFaceAngleYaw + 0x800
    if (o.oInteractStatus & INT_STATUS_INTERACTED) ~= 0 then
		spawn_object(o, E_MODEL_NONE, id_bhvGoldenCoinSparkles)
        obj_mark_for_deletion(o)
        o.oInteractStatus = 0
    end
end

hook_behavior(id_bhvSnowmansHead, OBJ_LIST_LEVEL, true, starsecret_init, starsecret_loop)

------ Fake Star ------

---@param o Object
local function bhv_fakestar_init(o)
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	obj_set_hitbox(o, sSecret)
	obj_set_model_extended(o, E_MODEL_STAR)
end

---@param o Object
local function bhv_fakestar_loop(o)
	o.oFaceAngleYaw = o.oFaceAngleYaw + 0x800
end

hook_behavior(id_bhvBigBoulder, OBJ_LIST_LEVEL, true, bhv_fakestar_init, bhv_fakestar_loop)

------ Fake Key ------

---@param o Object
local function fakekey_init(o)
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	obj_set_hitbox(o, sSecret)
	obj_set_model_extended(o, E_MODEL_BOWSER_KEY)
end

---@param o Object
local function fakekey_loop(o)
	o.oFaceAngleYaw = o.oFaceAngleYaw + 256
	if (o.oInteractStatus & INT_STATUS_INTERACTED) ~= 0 then
		spawn_object(o, E_MODEL_NONE, id_bhvGoldenCoinSparkles)
		obj_mark_for_deletion(o)
		o.oInteractStatus = 0
	end
end

hook_behavior(id_bhvSnowmansBottom, OBJ_LIST_LEVEL, true, fakekey_init, fakekey_loop)

------ Key Model for Star ------

---@param o Object
local function star_to_key_loop(o)
    local np = gNetworkPlayers[0]

    if np.currLevelNum == LEVEL_BOWSER_1 or np.currLevelNum == LEVEL_BOWSER_2 then
        obj_set_model_extended(o, E_MODEL_BOWSER_KEY)
        o.oFaceAngleRoll = -0x4000
        cur_obj_scale(0.4)
    end
end

hook_behavior(id_bhvStarSpawnCoordinates, OBJ_LIST_LEVEL, false, nil, star_to_key_loop)

------ Grand Star ------

--- @param o Object
local function gstar_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.oInteractType = INTERACT_WARP
    o.oIntangibleTimer = 0
    obj_scale(o, 5.3)
end

--- @param o Object
local function gstar_loop(o)
	o.oFaceAngleYaw = o.oFaceAngleYaw + 768
end

hook_behavior(id_bhvGrandStar, OBJ_LIST_LEVEL, true, gstar_init, gstar_loop)

------ Castle Star ------

--- @param o Object
local function finalstar_init(o)
	o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	bhv_init_room()
    bhv_collect_star_init()
	obj_scale(o, 2.4)
end

local function finalstar_loop()
	bhv_collect_star_loop()
end

hook_behavior(id_bhvHiddenAt120Stars, OBJ_LIST_LEVEL, true, finalstar_init, finalstar_loop)